function [BB, cv] = all_MAVE(x, y, method, CV);
%
% [estB,cv] = MAVE(x, y, method, CV);
% INPUT 
% x: nxp matrix of predictors
% y: nx1 of response;
% method: 'meanOPG'; 'meanMAVE'; 'csMAVE'; 'csOPG'; 'kSIR'
%         by default, method = 'csOPG' 
%         (1) 'meanOPG' and 'meanMAVE' estimate dimension reduction space 
%             for conditional mean
%         (2) 'csMAVE' and 'csOPG' estimate the central dimension reduction
%             space 
%         (3) 'kSIR' is a kernel version of SIR (Li, 1991). It is fast, but 
%             with poor efficiency 
% CV:  'yes'--to calculate the CV values
%      'no'--dont calculate CV values to make the calculation faster, by default 
% OUTPUT
% estB: pxpxp matrix
%     estB(:,1:d, d) is the central space with dimension=d
%         d = 1, 2, ..., p
% cv: Cross-Validation (CV) values for d = 1, 2, ..., p.
%     the dimension with smallest CV is the estimated dimension. 
%
% Reference,  Li K C (1991, JASA); Xia et al (2002, JRSSB); Xia (2007, AoS), Wang and Xia (2008, JASA)
%
% %Example
%  n = 200; p = 10;
%  B = zeros(p,2);
%  B(1,1)=1;
%  B(2,2)=1;
%  x = randn(n,p); 
%  y = (x*B(:,1)) + (x*B(:,2)).*randn(n,1);
%  [estB, cv] =all_MAVE(x,y,'csOPG', 'yes');
%  %Estimated directions 
%    B1 = estB(:,1:1,1)  % when dimension is fixed as 1
%    B2 = estB(:,1:2,2)  % when dimension is fixed as 2
%    B3 = estB(:,1:3,3)  % when dimension is fixed as 3
%    ...
%  % Estimated dimension
%     find(cv == min(cv))

[n,p] = size(x);
x = x - repmat(mean(x),n,1);
ss = pinv(x'*x/n + eye(p)/n^3)^0.5;
x = x*ss;

if nargin < 4
   CV = 'NO';
else
   CV = upper(CV);   
end

if nargin < 3
    disp('No method is specified, csOPG is used')
    method = 'CSOPG';
end
method = upper(method);
if ~(strcmp(method, 'CSOPG')|strcmp(method, 'CSMAVE')|strcmp(method, 'MEANOPG')|strcmp(method, 'MEANMAVE')|strcmp(method, 'KSIR'))
    warning('No such method, it must be one of: meanOPG; meanMAVE; csMAVE; csOPG; kSIR')
    disp('No estimation is done')
    return
end


if strcmp(method, 'MEANMAVE')|strcmp(method, 'MEANOPG')
    ky = y;
    ky1 = ky;
    ky2 = 1;
    DD = 1;    
else
    pp = max(min(99, n), floor(n^0.6));  
    qx = quantile(y, (1:pp)/(pp+1));

    qx = unique(qx);
    pp = length(qx);
    yi = repmat(y, 1, pp); 
    yj = repmat(qx, n,1);
    ky = abs(yi-yj<0);
    meanky = mean(ky);    
    ky = ky - repmat(meanky,n,1);
    ky1  = ky - repmat(mean(ky,2), 1, size(ky, 2));  
    ky1 = ky1 - repmat(mean(ky1),n,1);               
    ky = [ky ky1];                                   
    clear yi ky1;

    C = ky'*ky;                                     
    [V,D] = eig(C);                                 
    clear C;                                        
    [d,ix] = sort(abs(diag(D)),'descend');          
    di = min(find(cumsum(d)/sum(d) >= 0.99));       
    V = V(:,ix);                                    
    d = d(1:di);                                    
    V = V(:,1:di);                                  
    U = ky*V;                                       
    s = sqrt(d);                                    
    ky1 = bsxfun(@(x,c)x./c, U, s');                
    ky2 = diag(s)*V';                               
    ky = ky1*ky2;
    clear U V;
    DD = ky2*ky2';
    
    
end


if p == 0;
    BB = 1;
    cv = mean(mean(abs(ky - repmat(mean(ky), size(ky, 1), 1))));
    return 
end

BB = zeros(p,p,p);
BB(:,:,p) = eye(p);
onexi = ones(n,p+1);
m = p;
B = eye(p);

if p > 1

for ip = p:-1:1

    m = min(ip, floor(n/2));
    K = (ip<=10)*5 + (ip<=20)*(ip>10)*2 + (ip>20)*1;
    if strcmp(method, 'KSIR')
        K = 1;
    end

    for iter = 1:K;        
         
        if strcmp(method,'KSIR')
            V = y; m = 1;
        else
            V = x*B;
            npoints = max(min(100, n), floor(n^0.5));   
            Ifast = unifD(V, npoints);    
            Vfast = V(Ifast,:);         
            xfast = x(Ifast,:);         
            nfast = size(xfast, 1);     
        end
        h = 1.2*mean(std(V,[],1))/n^(1/(m+4));
        h2 = 2*h*h;
        h2i = h2;
        
        
        if strcmp(method,'CSOPG')|strcmp(method,'KSIR')|strcmp(method,'MEANOPG')|(ip>5)
            M0 = 0;
            for j = 1:nfast;           
                for i = 1:p
                    xij(:,i) = x(:,i) - xfast(j,i);
                end
                dxij = 0;
                for i = 1:m
                    dxij = dxij + (V(:,i)- Vfast(j,i)).^2;
                end                
                ds = sort(dxij);               
                h2i = max(h2, ds(2*m));
                ker = exp(-dxij/h2i);      
                onexi(:,1:p) = xij;    
                xk = (onexi.*repmat(ker, 1, p+1))';
                abi = inv(xk*onexi+eye(p+1)/n)*(xk*ky1);
                M0 = M0 + abi(1:p,:)*DD*abi(1:p,:)';                
                
            end	
            [B, R] = eig(M0);
            R = diag(R);
            [R, Ii] = sort(R, 'descend');
            Ball = B(:,Ii);
            B = Ball(:,1:ip);
        else
            dd = zeros(m*p, m*p);
            dc = zeros(m*p,1);
            D = 0;
            for j = 1:nfast;
                
                xij = x - repmat(xfast(j,:),n,1);   
                dxij = sum((V - repmat(Vfast(j,:),n,1)).^2,2);  
                ds = sort(dxij);
                h2i = max(h2, ds(2*m));
                ker = exp(-dxij/h2i);      
                rker = repmat(ker, 1, p+1);
           		onexi = [xij*B ones(n,1)];
                xk = (onexi.*rker(:,1:size(B,2)+1))';
                abi = inv(xk*onexi+eye(size(B,2)+1)/n)*(xk*ky1)*ky2;   
                kxij = (xij.*rker(:,1:p))';
                kxijy = kxij*(ky - repmat(abi(m+1,:),n,1));
                ddx = kxij*xij;
                for k1 = 1:m
                    ka = (k1-1)*p+1:k1*p;
                    dc(ka) = dc(ka) + kxijy*abi(k1,:)';
                end
                tmp = abi(1:m, :)*abi(1:m, :)';
                dd = dd + kron(tmp, ddx);
                D = D + tmp;
                
            end
    	   B = inv(dd + eye(length(dc))/n)*dc;
       	   B0 = reshape(B, p, m);
           [B, R] = eig(B0*D*B0');
           R = diag(R);
           [R, Ii] = sort(R, 'descend');
           Ball = B(:,Ii);
           B = Ball(:,1:ip);
        end
    end
    
    BB(:,1:ip, ip) = B;    

    if strcmp(method, 'KSIR')
        for jp = (p-1):-1:1
            BB(:,1:jp, jp) = Ball(:,1:jp);           
        end
        break        
    end
    
    B = B(:,1:(ip-1));
end

end

for ip = 1:p
    if strcmp(CV,'YES')
        if ip <= 5
            cv(ip) = CVm(x*BB(:,1:ip, ip), ky);
        else
            cv(ip) = cv(ip-1)+1;
        end
    end
    B = ss*BB(:,:,ip);
    for i = 1:size(B,2);
        B(:,i) = B(:,i)/(norm(B(:,i))+1.0e-20);
    end
    BB(:,:,ip) = B;
end

if strcmp(CV,'NO')
     warning('CV is not calculated')
     cv = 'CV is not calculated';
end

return


function  I = unifD(x, m);

% m is the number of points where teh gradients are calculated

m = floor(m);
[n,p] = size(x);
if ceil(n/m) == 1;
    I = 1:n;
    return
end

k = floor(n/m);
X = zeros(m,p);
Xremain = [x, (1:n)'];

I = [];
while size(Xremain,1) > 1
    mi = size(Xremain,1);
    xi = Xremain(1,1:p);
    di = sum((Xremain(:,1:p) - repmat(xi, mi, 1)).^2, 2);
    [di, Ii] = sort(di);
    Ii = Ii(1:min(k, mi));
    xi = mean(Xremain(Ii,1:p),1);
    [~,J] = sort(sum((Xremain(Ii,1:p) - repmat(xi, length(Ii), 1)).^2,2));
    I = [I, Xremain(Ii(J(1)),p+1)];
    Xremain(Ii,:)= [];
end



function cv = CVm(x, y)          

[n,p] = size(x);

x = tiedrank(x)/(n+1);    
h = mean(std(x))/n^(1/(p+3));
cv = 0;
for i = 1:n
        dxi = sum(abs(x - repmat(x(i,:),n,1)),2)/h;
        K = 1./(1+dxi).^4;
        K(i) = 0;
        ye = K'*y/sum(K);
        cv = cv + mean(mean(abs(ye-y(i,:))))/n;
end    
    

function cv = CVm31(x, y)

% 3-fold CV with random splitting of 100 times

[n,p] = size(x);
ny = size(y,2);
x = tiedrank(x)/(n+1);    

h = 1*mean(std(x,[],1))/n^(1/(p+3));
h2 = 2*h*h;

cv = 0;  
na = floor(n*2/3); nb = n-na;
J = 100;
dx = zeros(n);
for i = 1:n
%   dx(:,i) = sum((x - repmat(x(i,:),n,1)).^2,2)/h2;
     dx(:,i) = sum(abs(x - repmat(x(i,:),n,1)),2)/h;
end
for j = 1:J;
        rand('state', j); 
        [temp, I] = sort(rand(n,1));
        Ia = I(1:na);  Ib = I((na+1):n);        
        xA = x(Ia,:); yA = y(Ia,:);         
        xB = x(Ib,:); yB = y(Ib,:);        
        dxj = dx(Ia, Ib);
        ker = 1./(1+dxj).^4;
        ye = ker'*yA./repmat(sum(ker)'+1.0e-6, 1, ny);
        cv = cv + mean(mean(abs(yB-ye))); 
end    
    